#!/usr/bin/perl

package Text::Smudge;

use strict;
use File::Basename;
use File::Spec;
use File::Find;
use File::Path;
use Config::Tiny;
use Template;
use Fcntl ':mode';
use Params::Validate qw(validate SCALAR ARRAYREF UNDEF);

my $VERSION = '0.01';

# -------------------------------------------------------------------------
# Private methods

# Load data from given data files
sub _load_data
{
  my $self = shift;

  my %data = ();
  my $data_ts = 0;
  for my $data (@{$self->{data}}) {
    # Load data file
    unless (-r $data) {
      warn "missing datafile $data";
      next;
    }
    my $ts = (stat $data)[9];
    $data_ts = $ts if $ts > $data_ts;
    print "ts for data $data: $data_ts\n" if $self->{debug};
    my $Config = Config::Tiny->read($data);
    # @data{ keys %{$Config->{_}} } = values %{$Config->{_}} if keys %{$Config->{_}};
    for (keys %{$Config->{_}}) {
      # Eval [] and {} strings into arrayref and hashrefs respectively
      if (defined $Config->{_}->{$_} && $Config->{_}->{$_} =~ m/^[[{].*[\]}]$/) {
        unless ($data{$_} = eval "$Config->{_}->{$_}") {
          warn "Error: attempt to eval $_ reference failed!\n";
        }
      }
      # Otherwise just store data as a string
      else {
        $data{$_} = $Config->{_}->{$_};
      }
    }
  }
  print "newest data ts: $data_ts\n" if $self->{debug};
  $self->{data_ts} = $data_ts;
  $self->{data_hash} = \%data;
}

# Setup exclude_re arrayref for regexes and exclude_map hashref for strings
sub _setup_excludes
{
  my $self = shift;

  for (@{$self->{exclude}}) {
    if (m/(?<!\\)\*/) {
      s/(?<!\\)([.+?])/\\$1/g;
      s/\*/.*/g;
      push @{$self->{exclude_re}}, qr/$_/;
    }
    else {
      $self->{exclude_map}->{$_} = 1;
    }
  }
  print "exclude_re: " . join(',', @{$self->{exclude_re}}) . "\n" 
    if $self->{debug};
}

# Check whether the given filename has a template suffix, and remove it, if so.
# Return the filename (possibly truncated) and the tmpl index, if found.
sub _detemplatise
{
  my $self = shift;
  my ($filename) = @_;
  my $tmpl_index = undef;
  for (my $i = 0; $i <= $#{$self->{tmpl}}; $i++) {
    my $t = $self->{tmpl}->[$i];
    next unless m/\Q$t\E$/;
    # Match - truncate suffix and finish up
    $filename =~ s/\Q$t\E$//;
    $tmpl_index = $i;
    last;
  }
  return wantarray ? ($filename, $tmpl_index) : $filename;
}

# Setup/cleanup $outfile
sub _map_outfile
{
  my $self = shift;
  my ($file, $outfile, $dir) = @_;
  $outfile = sprintf $self->{format}, $outfile if $self->{format};
  if ($self->{outdir}) {
    my $finddir = File::Spec->abs2rel($dir, $self->{tree});
    # Detemplatise (i.e. rename) any directory components
    $finddir = File::Spec->catdir(map { scalar $self->_detemplatise($_) } 
      File::Spec->splitdir($finddir));
    my $od = File::Spec->catdir($self->{outdir}, $finddir);
    unless (-d $od) {
      mkpath($od) || die "Could not create directory $od: $!\n";
    }
    $outfile = File::Spec->catfile($od, $outfile);
  }
  return $outfile;
}

# Return true if $outfile exists and is newer than template and data
sub _is_newer
{
  my $self = shift;
  my ($file, $outfile, $tmpl_index) = @_;

  my $tmpl_ts = (stat $file)[9];
  printf "ts for tmpl %s: %s\n", $file, $tmpl_ts if $self->{debug};
  
  # Exception - if outfile was produced this run by a lower order template
  if (-f $outfile && ! $self->{force} &&
      (! exists $self->{smudged}->{$outfile} || 
                $self->{smudged}->{$outfile} > $tmpl_index)) {
    my $outfile_ts = (stat $outfile)[9];
    print "ts for outfile: $outfile_ts\n" if $self->{debug};
    return 1 if $outfile_ts > $self->{data_ts} && $outfile_ts > $tmpl_ts;
    print "outfile out of date - continuing\n" if $self->{debug};
  }
  return 0;
}

sub _add_smudge_vars
{
  my $self = shift;
  my ($file) = @_;
  my $tpath = File::Spec->rel2abs($file);
  $self->{data_hash}->{SMUDGE_BASENAME} = basename $tpath;
  $self->{data_hash}->{SMUDGE_DIRNAME} = dirname $tpath;
  $self->{data_hash}->{SMUDGE_PARENT} = basename dirname $tpath;
  $self->{data_hash}->{SMUDGE_DIRNAME2} = dirname dirname $tpath;
  $self->{data_hash}->{SMUDGE_PARENT2} = basename dirname dirname $tpath;
  $self->{data_hash}->{SMUDGE_DIRNAME3} = dirname dirname dirname $tpath;
  $self->{data_hash}->{SMUDGE_PARENT3} = basename dirname dirname dirname $tpath;
}

# Check for variables in the templates with undefined values - report and abort
sub _check_unsmudged
{
  my $self = shift;
  my ($unsmudged, $outfile, $infile);
  if (keys %$unsmudged) {
    unlink $outfile;
    for my $var (sort keys %$unsmudged) {
      print STDERR "Undefined variable $var occurs in $infile\n";
    }
    die "Undefined smudge variables found - aborting\n";
  }
}

# Do the merging
sub _smudge_file {
  my $self = shift;
  my ($file, $dir, $name) = @_;

  return unless -f $file;

  my ($outfile, $tmpl_index) = $self->_detemplatise($file);
  return unless defined $tmpl_index;

  # Setup outfile
  $outfile = $self->_map_outfile($file, $outfile, $dir);
  print "outfile for $file: $outfile\n" if $self->{debug};

  # Skip iff outfile exists, and is newer than both data and template
  return if $self->_is_newer($file, $outfile, $tmpl_index);

  # Add some internal SMUDGE variables to $self->{data_hash}
  $self->_add_smudge_vars($file);

  # Relativise paths for reporting
  my $infile_rel = File::Spec->abs2rel($name, $self->{tree});
  my $outfile_rel = File::Spec->abs2rel($outfile, $self->{tree});
  printf "o %-30s -> %s\n", $infile_rel, $outfile_rel if $self->{verbose};

  # Initialise TT object and do main TT processing
  my $stash = Template::SmudgeStash->new;
  my $tt = Template->new({ STASH => $stash, EVAL_PERL => 1 });
  $tt->process($file, $self->{data_hash}, $outfile) || die $tt->error;

  # Check for (abort on) template variables with undefined values
  $self->_check_unsmudged($stash->unsmudged, $outfile, $infile_rel);

  # Preserve outfile permissions
  chmod S_IMODE((stat $name)[2]), $outfile;

  # All good - record outfile and tmpl_index, and return
  $self->{smudged}->{$outfile} = $tmpl_index;
}

# Exclude undesirables from the given filename list
sub _exclude 
{
  my $self = shift;
  my @include;
  FILE: for my $x (@_) {
    # Template suffixes override excludes
    for my $t (@{$self->{tmpl}}) {
      if ($x =~ m/\Q$t\E$/) {
        push @include, $x;
        next FILE;
      }
    }
    # Test excludes
    next FILE if $self->{exclude_map}->{$x};
    if (@{$self->{exclude_re}}) {
      for my $re (@{$self->{exclude_re}}) {
        next FILE if $x =~ $re;
      }
    }
    # Keep directories to recurse into
    push @include, $x if -d $x;
  }
  return @include; 
}

# Walk the given tree, doing the merges
sub _process
{
  my $self = shift;
  find({ 
    wanted => sub {
      $self->_smudge_file($_, $File::Find::dir, $File::Find::name)
    }, 
    @{$self->{exclude}} ? ( preprocess => sub { 
      $self->_exclude(@_)
    } ) : () 
  }, $self->{tree});
}

# -------------------------------------------------------------------------
# Constructor
sub new
{
  my $class = shift;
  my %arg = validate( @_, {
    tree => { type => SCALAR, default => '.', },
    outdir => { type => SCALAR, default => '.', },
    tmpl => { type => SCALAR | ARRAYREF, default => '.tmpl', },
    data => { type => SCALAR | ARRAYREF, },
    format => { type => SCALAR, optional => 1, },
    exclude => { type => SCALAR | ARRAYREF, optional => 1, },
    force => { type => SCALAR, optional => 1, },
    verbose => { type => SCALAR, optional => 1, },
    debug => { type => SCALAR, optional => 1, },
  });
  my $self = bless \%arg, $class;

  # Setup/cleanup path args
  $self->{outdir} = undef if $self->{outdir} eq '.';
  $self->{format} ||= '%s.out' if ! $self->{outdir};
  $self->{tree} = File::Spec->curdir if $self->{tree} eq '.';
  $self->{tree} = File::Spec->rel2abs($self->{tree});

  # Force arrayref args
  $self->{data} = [ $self->{data} ] unless ref $self->{data};
  $self->{tmpl} = [ $self->{tmpl} ] unless ref $self->{tmpl};
  $self->{exclude} = [ $self->{exclude} ] unless ref $self->{exclude};

  # Absolutise outdir and create, if necessary
  if ($self->{outdir}) {
    $self->{outdir} = File::Spec->rel2abs($self->{outdir});
    mkpath($self->{outdir}) unless -d $self->{outdir};
    printf "outdir: %s\n", $self->{outdir} if $self->{debug};
  }

  $self->{exclude_re} = [];
  $self->{exclude_map} = {};
  $self->{smudged} = {};

  return $self;
}

# -------------------------------------------------------------------------
# Main class method

sub smudge
{
  my $class = shift;
  my $self = ref $class ? $class : $class->new(@_);

  # Load data from data files
  $self->_load_data;

  # Setup exclusions from exclude args
  $self->_setup_excludes;

  # Process the input tree
  $self->_process;
}

# -------------------------------------------------------------------------
# Template::SmudgeStash
# -------------------------------------------------------------------------

# Subclass of Template::Stash to track undefined variables
package Template::SmudgeStash;

use Template::Stash;

use vars qw(@ISA);
@ISA = qw(Template::Stash);

my @IGNORE = qw(component);

sub new
{
  my $this = shift;
  my $class = ref $this || $this; 
  my $self = $this->SUPER::new(@_);
  bless $self, $this;
  $self->{_smudge} = {};
  $self
}

sub undefined { undef };

#
# Subclass get to record variables without a value
#
sub get 
{
  my $self = shift;
  my ($var) = @_;
  my $value = $self->SUPER::get(@_);
  return $value if grep /^$var$/, @IGNORE;
  $self->{_smudge}->{$var} = 1 if ! defined $value;
  return $value;
}

#
# Return a hashref of the empty variables we've seen
#
sub unsmudged
{
  my $self = shift;
  return $self->{_smudge};
}

# -------------------------------------------------------------------------

1;

__END__

=head1 NAME

Text::Smudge is a perl module for merging a set of data into multiple
templates within a directory tree, producing a new merged output file
for every template it finds


=head1 SYNOPSIS

    # Generate an 'out' tree from templates in 'tree' using the given data
    Text::Smudge->smudge(
        tree => 'tree',
        outdir => 'out',
        tmpl => [ qw(.tmpl .tmpl.web .tmpl.centos) ],
        data => [ qw(global.conf web.conf centos.conf web1.conf) ],
    );


=head1 DESCRIPTION

smudge is a tool for merging a set of data into multiple templates 
within a directory tree, producing a new merged output file for every 
template it finds. It was motivated by the need to maintain config
trees (subsets of '/etc') for multiple similar hosts (e.g. 
load-balancing webservers), where the configs are almost identical
except for details like ip addresses etc.

smudge requires one or more data files containing NAME=value pairs, one 
per line, defining the variables you want to substitute in your 
templates; a set of templates containing the merge fields you want to 
have substituted, identified by a filename suffix; optionally, an 
'outdir' directory to use for the output file tree to contain the 
smudged output files; optionally, a 'format' argument defining the 
format of the filenames of the merge output files produced; and the 
tree within which to search for templates. See the L<ARGUMENTS> section 
below for more details on each of these.

=head2 ARGUMENTS

=over 4

=item tree => <tree>

Path to the tree within which to look for templates. Default: '.' 
i.e. the current directory.

=item outdir => <outdir>

Directory within which output files should be saved. Directory structure
relative to <tree> is preserved within <outdir> i.e. a template like
<tree>/etc/httpd/conf/httpd.conf.tmpl would be saved as
<outdir>/etc/httpd/conf/<outfilename> (where <outfilename> is 
determined by <format> below). If outdir is not set, merged output
files are generated in situ in <tree>.

=item format => <format>

The output format is used in constructing the filename of the merged 
output file. The template suffix <tsuff> is first removed from the 
filename, and then <format> is used as a sprintf format to generate the 
filename. So, for instance, given the default template suffix '.tmpl' 
and a template called 'network.tmpl':

  -f '%.out' 

will produce an output file called 'network.out'; and:

  -f '++%s.web1.out'
 
will product an output file called '++network.web1.out'.

Default: if <outdir> is set, the default <format> is '%s' (i.e. the 
filename without the template suffix); if <outdir> is not set, the 
default <format> is '%.out'.


=item tmpl => <tsuff>

Scalar or arrayref of template suffixes to search for within the tree. 
Suffixes are always removed before constructing the output filename. 
Default: '.tmpl'.

Templates should be text files containing tags or placeholders to be 
replaced using the datafile data at merge time. Tags should be formatted:

  [% NAME %]

e.g. resolv.conf.tmpl:

  domain openfusion.com.au
  nameserver [% NS1 %]
  nameserver [% NS2 %]

where the [% NS1 %] and [% NS2 %] tags will be replaced by the value of 
the NS1 and NS2 variables defined in the data file.

Where multiple template suffixes are used via an arrayref e.g.

  tmpl => [ qw(.tmpl .tmpl.web .tmpl.centos) ]

then later templates take precedence over earlier ones i.e. given the
above template list and two templates, 'foo.conf.tmpl' and 
'foo.conf.tmpl.centos', the second of those would be used to 
generate a foo.conf output file.

=item data => <data>

Mandatory scalar or arrayref, defining paths to one or more data files 
containing the variables (NAME=value pairs, one per line) to be substituted 
into templates. The data may contain comment lines beginning with '#', and 
blank lines are ignored. 

Where multiple data files are supplied via an arrayref e.g.

  data => [ qw(global.conf web.conf centos.conf host1.conf) ]

they are merged in the order given i.e. variables with the same name in 
later data files take precedence over earlier ones.

=item exclude => <exclude>

Scalar or arrayref argument defining items to exclude in our traverse 
of the template tree. Excludes may be simple strings or filename globs
(not full regexes). Multiple excludes may be specified using an arrayref
e.g.

  exclude => [ qw(CVS {arch} .svn *~ *.bak) ]
  
Note that template patterns take precedence over excludes, so that you
can exclude template patterns safely if you have templates you want to
ignore unless explicitly used e.g. given the following templates:

  HOSTNAME.tmpl
  sysconfig/network.tmpl.centos
  sysconfig/network.tmpl.suse

using the following arguments:

  tmpl => [ qw(.tmpl .tmpl.centos) ],
  exclude => '*.tmpl.*'

would exclude the 'suse' template, but still use the other two.

=item force => 1

Force output to always be updated, ignoring normal timestamp checks.


=head2 SPECIAL SMUDGE VARIABLES

smudge sets the following special smudge variables that may be accessed
within templates as normal e.g.

  [% SMUDGE_BASENAME %]

The current smudge variables are:

=over 4

=item SMUDGE_BASENAME

The basename (bare filename) of the current template.

=item SMUDGE_DIRNAME

The dirname (absolute path of the directory) of the current template.

=item SMUDGE_PARENT

The final component of the directory of the current template (i.e.
basename dirname TEMPLATE).

=item SMUDGE_DIRNAME2, SMUDGE_DIRNAME3

The dirnames of the parent and grandparent directories of the current 
template.

=item SMUDGE_PARENT2, SMUDGE_PARENT3

The final component (i.e. basename) of the directories of the parent and 
grandparent directories of the current template.

=back


=head1 DIRECTORY SMUDGING

As of version 0.7, smudge also supports directory smudging, which means
it will rename directories that match the given template patterns in 
the same way it renames output files. This allows alternative directory
trees to be supported. For instance, given:

  magic.tmpl.suse        (directory)
  +-- config1.tmpl
  `-- config2.tmpl
  magic.tmpl.centos      (directory)
  +-- completely-different-config1.tmpl
  `-- completely-different-config2.tmpl

specifying templates and excludes as follows:

  -t .tmpl -t .tmpl.suse -x *.tmpl.*

and 

  -t .tmpl -t .tmpl.centos -x *.tmpl.*

would produce on SuSE:

  magic
  +-- config1 
  `-- config2 

and on CentOS:

  magic
  +-- completely-different-config1
  +-- completely-different-config2

(assuming a 'format' of '%s').


=head1 MODULES

Requires the following non-core perl modules, available from your 
local CPAN:

=over 4

=item Config::Tiny

=item Template

=back 


=head1 SEE ALSO

Template Toolkit includes a utility called 'ttree' which is a more generic 
version of smudge, exposing more of TT's complexities. Smudge is simpler,
more focussed, and hopefully easier to learn and use.


=head1 AUTHOR

Gavin Carr <gavin@openfusion.com.au>


=head1 COPYRIGHT AND LICENCE

Copyright 2004-2006 Gavin Carr <gavin@openfusion.com.au>. All Rights 
Reserved.

This program is free software licensed under the GNU GPL. See the file
COPYING for details.

=cut

# arch-tag: 0914b296-d316-449a-845c-bb1951344523
